@@ -1,5 +1,7 @@
 {-# LANGUAGE CPP                #-}
+{-# LANGUAGE DeriveGeneric      #-}
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE TemplateHaskell    #-}
 
 -- | Internal types and functions.
 
@@ -12,10 +14,14 @@ module Path.Internal
   where
 
 import Control.DeepSeq (NFData (..))
-import Data.Aeson (ToJSON (..))
+import Data.Aeson (ToJSON (..), ToJSONKey(..))
+import Data.Aeson.Types (toJSONKeyText)
+import qualified Data.Text as T (pack)
+import GHC.Generics (Generic)
 import Data.Data
 import Data.Hashable
 import Data.List
+import Language.Haskell.TH.Syntax (Exp(..), Lift(..), Lit(..))
 import qualified System.FilePath as FilePath
 
 -- | Path of some base and type.
@@ -33,7 +39,7 @@ import qualified System.FilePath as FilePath
 -- All directories end in a trailing separator. There are no duplicate
 -- path separators @\/\/@, no @..@, no @.\/@, no @~\/@, etc.
 newtype Path b t = Path FilePath
-  deriving (Typeable)
+  deriving (Data, Typeable, Generic)
 
 -- | String equality.
 --
@@ -83,6 +89,9 @@ instance ToJSON (Path b t) where
   {-# INLINE toEncoding #-}
 #endif
 
+instance ToJSONKey (Path b t) where
+  toJSONKey = toJSONKeyText $ T.pack . toFilePath
+
 instance Hashable (Path b t) where
   -- A "." is represented as an empty string ("") internally. Hashing ""
   -- results in a hash that is the same as the salt. To produce a more
@@ -103,3 +112,6 @@ hasParentDir filepath' =
         case FilePath.pathSeparator of
             '/' -> filepath'
             x   -> map (\y -> if x == y then '/' else y) filepath'
+
+instance Lift (Path a b) where
+  lift (Path str) = [|Path $(return (LitE (StringL str)))|]
