@@ -11,11 +11,14 @@ import Control.Applicative
 import Control.Monad
 import Data.Aeson
 import qualified Data.ByteString.Lazy.Char8 as LBS
+import Data.Function (on)
 import Data.Maybe
-import Path
+import Path.Windows
 import Path.Internal
 import Test.Hspec
 
+import Common (extensionOperations)
+
 -- | Test suite (Windows version).
 spec :: Spec
 spec =
@@ -30,8 +33,7 @@ spec =
      describe "Operations: parent" operationParent
      describe "Operations: filename" operationFilename
      describe "Operations: dirname" operationDirname
-     describe "Operations: addFileExtension" operationAddFileExtension
-     describe "Operations: setFileExtension" operationSetFileExtension
+     describe "Operations: extensions" (extensionOperations "C:\\")
      describe "Restrictions" restrictions
      describe "Aeson Instances" aesonInstances
      describe "QuasiQuotes" quasiquotes
@@ -57,43 +59,57 @@ restrictions =
 
 -- | The 'dirname' operation.
 operationDirname :: Spec
-operationDirname = do
-  it
-    "dirname ($(mkAbsDir parent) </> $(mkRelFile dirname)) == dirname $(mkRelFile dirname) (unit test)"
-    (dirname ($(mkAbsDir "C:\\chris\\") </> $(mkRelDir "bar")) ==
-     dirname $(mkRelDir "bar"))
-  it
-    "dirname ($(mkRelDir parent) </> $(mkRelFile dirname)) == dirname $(mkRelFile dirname) (unit test)"
-    (dirname ($(mkRelDir "home\\chris\\") </> $(mkRelDir "bar")) ==
-     dirname $(mkRelDir "bar"))
-  it
-    "dirname $(mkRelDir .) == $(mkRelDir .)"
-    (dirname $(mkRelDir ".") == $(mkRelDir "."))
-  it
-    "dirname C:\\ must be a Rel path"
-    ((parseAbsDir $ show $ dirname (fromJust (parseAbsDir "C:\\"))
-     :: Maybe (Path Abs Dir)) == Nothing)
+operationDirname =
+  do it "dirname ($(mkAbsDir parent) </> $(mkRelFile dirname)) == dirname $(mkRelFile dirname) (absolute)"
+        (dirnamesShouldBeEqual
+          ($(mkAbsDir "C:\\chris\\") </> $(mkRelDir "bar"))
+          $(mkRelDir "bar"))
+     it "dirname ($(mkRelDir parent) </> $(mkRelFile dirname)) == dirname $(mkRelFile dirname) (relative)"
+        (dirnamesShouldBeEqual
+          ($(mkRelDir "home\\chris\\") </> $(mkRelDir "bar"))
+          $(mkRelDir "bar"))
+     it "dirname ($(mkAbsDir parent) </> $(mkRelFile dirname)) == dirname $(mkRelFile dirname) (UNC)"
+        (dirnamesShouldBeEqual
+          ($(mkAbsDir "\\\\home\\chris\\") </> $(mkRelDir "bar"))
+          $(mkRelDir "bar"))
+     it "dirname ($(mkAbsDir parent) </> $(mkRelFile dirname)) == dirname $(mkRelFile dirname) (Unicode)"
+        (dirnamesShouldBeEqual
+          ($(mkAbsDir "\\\\?\\C:\\home\\chris\\") </> $(mkRelDir "bar"))
+          $(mkRelDir "bar"))
+     it "dirname $(mkRelDir .) == $(mkRelDir .)"
+        (dirnamesShouldBeEqual
+          $(mkRelDir ".")
+          $(mkRelDir "."))
+     it "dirname C:\\ must be a Rel path"
+        ((parseAbsDir $ show $ dirname (fromJust (parseAbsDir "C:\\")) :: Maybe (Path Abs Dir)) == Nothing)
+  where dirnamesShouldBeEqual = (==) `on` dirname
 
 -- | The 'filename' operation.
 operationFilename :: Spec
 operationFilename =
-  do it "filename ($(mkAbsDir parent) </> $(mkRelFile filename)) == filename $(mkRelFile filename) (unit test)"
-          (filename ($(mkAbsDir "C:\\chris\\") </>
-                             $(mkRelFile "bar.txt")) ==
-                                      filename $(mkRelFile "bar.txt"))
-
-     it "filename ($(mkRelDir parent) </> $(mkRelFile filename)) == filename $(mkRelFile filename) (unit test)"
-             (filename ($(mkRelDir "home\\chris\\") </>
-                                $(mkRelFile "bar.txt")) ==
-                                         filename $(mkRelFile "bar.txt"))
+  do it "filename ($(mkAbsDir parent) </> $(mkRelFile filename)) == filename $(mkRelFile filename) (absolute)"
+        (filenamesShouldBeEqual
+          ($(mkAbsDir "C:\\chris\\") </> $(mkRelFile "bar.txt"))
+          $(mkRelFile "bar.txt"))
+     it "filename ($(mkRelDir parent) </> $(mkRelFile filename)) == filename $(mkRelFile filename) (relative)"
+        (filenamesShouldBeEqual
+          ($(mkRelDir "home\\chris\\") </> $(mkRelFile "bar.txt"))
+          $(mkRelFile "bar.txt"))
+     it "filename ($(mkAbsDir parent) </> $(mkRelFile filename)) == filename $(mkRelFile filename) (UNC)"
+        (filenamesShouldBeEqual
+          ($(mkAbsDir "\\\\host\\share\\chris\\") </> $(mkRelFile "bar.txt"))
+          $(mkRelFile "bar.txt"))
+     it "filename ($(mkAbsDir parent) </> $(mkRelFile filename)) == filename $(mkRelFile filename) (Unicode)"
+        (filenamesShouldBeEqual
+          ($(mkAbsDir "\\\\?\\C:\\home\\chris\\") </> $(mkRelFile "bar.txt"))
+          $(mkRelFile "bar.txt"))
+  where filenamesShouldBeEqual = (==) `on` filename
 
 -- | The 'parent' operation.
 operationParent :: Spec
 operationParent =
   do it "parent (parent </> child) == parent"
-        (parent ($(mkAbsDir "C:\\foo") </>
-                    $(mkRelDir "bar")) ==
-         $(mkAbsDir "C:\\foo"))
+        (parent ($(mkAbsDir "C:\\foo") </> $(mkRelDir "bar")) == $(mkAbsDir "C:\\foo"))
      it "parent \"C:\\\" == \"C:\\\""
         (parent $(mkAbsDir "C:\\") == $(mkAbsDir "C:\\"))
      it "parent \"C:\\x\" == \"C:\\\""
@@ -108,67 +124,103 @@ operationIsProperPrefixOf :: Spec
 operationIsProperPrefixOf =
   do it "isProperPrefixOf parent (parent </> child) (absolute)"
         (isProperPrefixOf
-           $(mkAbsDir "C:\\\\\\bar\\")
-           ($(mkAbsDir "C:\\\\\\bar\\") </>
-            $(mkRelFile "bar\\foo.txt")))
-
+          $(mkAbsDir "C:\\\\\\bar\\")
+          ($(mkAbsDir "C:\\\\\\bar\\") </> $(mkRelFile "bar\\foo.txt")))
      it "isProperPrefixOf parent (parent </> child) (relative)"
         (isProperPrefixOf
-           $(mkRelDir "bar\\")
-           ($(mkRelDir "bar\\") </>
-            $(mkRelFile "bob\\foo.txt")))
-
+          $(mkRelDir "bar\\")
+          ($(mkRelDir "bar\\") </> $(mkRelFile "bob\\foo.txt")))
+     it "isProperPrefixOf parent (parent </> child) (UNC)"
+        (isProperPrefixOf
+          $(mkAbsDir "\\\\host\\share\\")
+          ($(mkAbsDir "\\\\host\\share\\") </> $(mkRelFile "bob\\foo.txt")))
+     it "isProperPrefixOf parent (parent </> child) (Unicode)"
+        (isProperPrefixOf
+          $(mkAbsDir "\\\\?\\C:\\folder\\")
+          ($(mkAbsDir "\\\\?\\C:\\folder\\") </> $(mkRelFile "bob\\foo.txt")))
      it "not (x `isProperPrefixOf` x)"
         (not (isProperPrefixOf $(mkRelDir "x") $(mkRelDir "x")))
-
      it "not (\\ `isProperPrefixOf` \\)"
         (not (isProperPrefixOf $(mkAbsDir "C:\\") $(mkAbsDir "C:\\")))
 
 -- | The 'stripProperPrefix' operation.
 operationStripProperPrefix :: Spec
 operationStripProperPrefix =
-  do it "stripProperPrefix parent (parent </> child) = child (unit test)"
-        (stripProperPrefix $(mkAbsDir "C:\\\\\\bar\\")
-                  ($(mkAbsDir "C:\\\\\\bar\\") </>
-                   $(mkRelFile "bar\\foo.txt")) ==
-         Just $(mkRelFile "bar\\foo.txt"))
-
-     it "stripProperPrefix parent (parent </> child) = child (unit test)"
-        (stripProperPrefix $(mkRelDir "bar\\")
-                  ($(mkRelDir "bar\\") </>
-                   $(mkRelFile "bob\\foo.txt")) ==
-         Just $(mkRelFile "bob\\foo.txt"))
-
+  do it "stripProperPrefix parent (parent </> child) = child (absolute)"
+        (remainingPathShouldBe
+          $(mkAbsDir "C:\\\\\\bar\\")
+          ($(mkAbsDir "C:\\\\\\bar\\") </> $(mkRelFile "bar\\foo.txt"))
+          (Just $(mkRelFile "bar\\foo.txt")))
+     it "stripProperPrefix parent (parent </> child) = child (relative)"
+        (remainingPathShouldBe
+          $(mkRelDir "bar\\")
+          ($(mkRelDir "bar\\") </> $(mkRelFile "bob\\foo.txt"))
+          (Just $(mkRelFile "bob\\foo.txt")))
+     it "stripProperPrefix parent (parent </> child) = child (UNC)"
+        (remainingPathShouldBe
+          $(mkAbsDir "\\\\host\\share\\")
+          ($(mkAbsDir "\\\\host\\share\\") </> $(mkRelFile "bob\\foo.txt"))
+          (Just $(mkRelFile "bob\\foo.txt")))
+     it "stripProperPrefix parent (parent </> child) = child (Unicode)"
+        (remainingPathShouldBe
+          $(mkAbsDir "\\\\?\\C:\\folder\\")
+          ($(mkAbsDir "\\\\?\\C:\\folder\\") </> $(mkRelFile "bob\\foo.txt"))
+          (Just $(mkRelFile "bob\\foo.txt")))
      it "stripProperPrefix parent parent = _|_"
-        (stripProperPrefix $(mkAbsDir "C:\\home\\chris\\foo")
-                  $(mkAbsDir "C:\\home\\chris\\foo") ==
-         Nothing)
+        (remainingPathShouldBe
+          $(mkAbsDir "C:\\home\\chris\\foo")
+          $(mkAbsDir "C:\\home\\chris\\foo")
+          Nothing)
+  where remainingPathShouldBe prefix path suffix =
+          stripProperPrefix prefix path == suffix
 
 -- | The '</>' operation.
 operationAppend :: Spec
 operationAppend =
   do it "AbsDir + RelDir = AbsDir"
-        ($(mkAbsDir "C:\\home\\") </>
-         $(mkRelDir "chris") ==
-         $(mkAbsDir "C:\\home\\chris\\"))
+        (shouldBe
+          ($(mkAbsDir "C:\\home\\") </> $(mkRelDir "chris"))
+          $(mkAbsDir "C:\\home\\chris\\"))
      it "AbsDir + RelFile = AbsFile"
-        ($(mkAbsDir "C:\\home\\") </>
-         $(mkRelFile "chris\\test.txt") ==
-         $(mkAbsFile "C:\\home\\chris\\test.txt"))
+        (shouldBe
+          ($(mkAbsDir "C:\\home\\") </> $(mkRelFile "chris\\test.txt"))
+          $(mkAbsFile "C:\\home\\chris\\test.txt"))
      it "RelDir + RelDir = RelDir"
-        ($(mkRelDir "home\\") </>
-         $(mkRelDir "chris") ==
-         $(mkRelDir "home\\chris"))
+        (shouldBe
+          ($(mkRelDir "home\\") </> $(mkRelDir "chris"))
+          $(mkRelDir "home\\chris"))
      it ". + . = ."
-        ($(mkRelDir ".\\") </> $(mkRelDir ".") == $(mkRelDir "."))
+        (shouldBe
+          ($(mkRelDir ".\\") </> $(mkRelDir "."))
+          $(mkRelDir "."))
      it ". + x = x"
-        ($(mkRelDir ".") </> $(mkRelDir "x") == $(mkRelDir "x"))
+        (shouldBe
+          ($(mkRelDir ".") </> $(mkRelDir "x"))
+          $(mkRelDir "x"))
      it "x + . = x"
-        ($(mkRelDir "x") </> $(mkRelDir ".\\") == $(mkRelDir "x"))
+        (shouldBe
+          ($(mkRelDir "x") </> $(mkRelDir ".\\"))
+          $(mkRelDir "x"))
      it "RelDir + RelFile = RelFile"
-        ($(mkRelDir "home\\") </>
-         $(mkRelFile "chris\\test.txt") ==
-         $(mkRelFile "home\\chris\\test.txt"))
+        (shouldBe
+          ($(mkRelDir "home\\") </> $(mkRelFile "chris\\test.txt"))
+          $(mkRelFile "home\\chris\\test.txt"))
+     it "AbsDir(UNC) + RelDir = AbsDir(UNC)"
+        (shouldBe
+          ($(mkAbsDir "\\\\host\\share\\") </> $(mkRelDir "folder\\"))
+          $(mkAbsDir "\\\\host\\share\\folder\\"))
+     it "AbsDir(UNC) + RelFile = AbsFile(UNC)"
+        (shouldBe
+          ($(mkAbsDir "\\\\host\\share\\") </> $(mkRelFile "folder\\file.txt"))
+          $(mkAbsFile "\\\\host\\share\\folder\\file.txt"))
+     it "AbsDir(Unicode) + RelDir = AbsDir(Unicode)"
+        (shouldBe
+          ($(mkAbsDir "\\\\?\\C:\\folder\\") </> $(mkRelDir "another\\"))
+          $(mkAbsDir "\\\\?\\C:\\folder\\another\\"))
+     it "AbsDir(Unicode) + RelFile = AbsFile(Unicode)"
+        (shouldBe
+          ($(mkAbsDir "\\\\?\\C:\\folder\\") </> $(mkRelFile "file.txt"))
+          $(mkAbsFile "\\\\?\\C:\\folder\\file.txt"))
 
 operationToFilePath :: Spec
 operationToFilePath =
@@ -177,56 +229,24 @@ operationToFilePath =
      it "show $(mkRelDir \".\") == \"\\\".\\\\\"\""
         (show $(mkRelDir ".") == "\".\\\\\"")
 
-operationAddFileExtension :: Spec
-operationAddFileExtension = do
-  it "adds extension if there is none" $
-    addFileExtension "ext" $(mkAbsFile "C:\\directory\\path")
-      `shouldReturn` $(mkAbsFile "C:\\directory\\path.ext")
-  it "adds extension if there is already one" $
-    addFileExtension "baz" $(mkRelFile "foo.bar")
-      `shouldReturn` $(mkRelFile "foo.bar.baz")
-  it "adds extension with dot" $
-    addFileExtension ".bar" $(mkRelFile "foo")
-      `shouldReturn` $(mkRelFile "foo.bar")
-  it "adds extension with dot after dot" $
-    $(mkRelFile "foo.") <.> ".bar"
-      `shouldReturn` $(mkRelFile "foo..bar")
-  it "adds extension without dot after dot" $
-    $(mkRelFile "foo.") <.> "bar"
-      `shouldReturn` $(mkRelFile "foo..bar")
-  it "adds extension with separator" $  -- I'm not sure it's okay
-    $(mkRelFile "foo") <.> "evil\\extension"
-      `shouldReturn` $(mkRelFile "foo.evil\\extension")
-  it "adds extension to file inside dotted directory" $
-    $(mkRelFile "foo.bar\\baz") <.> "txt"
-      `shouldReturn` $(mkRelFile "foo.bar\\baz.txt")
-  it "throws InvalidRelFile extension if extenstion ends with \\" $
-    $(mkRelFile "foo") <.> "evil\\"
-      `shouldThrow` (== InvalidRelFile "foo.evil\\")
-  it "throws InvalidAbsFile extension if extenstion ends with \\" $
-    $(mkAbsFile "C:\\home\\cfg") <.> "txt\\"
-      `shouldThrow` (== InvalidAbsFile "C:\\home\\cfg.txt\\")
-
-operationSetFileExtension :: Spec
-operationSetFileExtension = do
-  it "adds extension if there is none" $
-    setFileExtension "txt" $(mkRelFile "foo")
-      `shouldReturn` $(mkRelFile "foo.txt")
-  it "replaces extension if the input path already has one" $
-    setFileExtension "txt" $(mkRelFile "foo.bar")
-      `shouldReturn` $(mkRelFile "foo.txt")
-
 -- | Tests for the tokenizer.
 parseAbsDirSpec :: Spec
 parseAbsDirSpec =
   do failing ""
      failing ".\\"
      failing "foo.txt"
+     failing "C:"
      succeeding "C:\\" (Path "C:\\")
-     succeeding "C:\\\\" (Path "C:\\\\")
-     -- succeeding "C:\\\\\\foo\\\\bar\\\\mu\\" (Path "C:\\foo\\bar\\mu\\") FIXME
-     -- succeeding "C:\\\\\\foo\\\\bar\\\\mu" (Path "C:\\foo\\bar\\mu\\") FIXME
-     -- succeeding "C:\\\\\\foo\\\\bar\\.\\\\mu" (Path "C:\\foo\\bar\\mu\\") FIXME
+     succeeding "C:\\\\" (Path "C:\\")
+     succeeding "C:\\\\\\foo\\\\bar\\\\mu\\" (Path "C:\\foo\\bar\\mu\\")
+     succeeding "C:\\\\\\foo\\\\bar\\\\mu" (Path "C:\\foo\\bar\\mu\\")
+     succeeding "C:\\\\\\foo\\\\bar\\.\\\\mu" (Path "C:\\foo\\bar\\mu\\")
+     succeeding "\\\\unchost\\share" (Path "\\\\unchost\\share\\")
+     succeeding "\\/unchost\\share" (Path "\\\\unchost\\share\\")
+     succeeding "\\\\unchost\\share\\\\folder\\" (Path "\\\\unchost\\share\\folder\\")
+     succeeding "\\\\?\\C:\\" (Path "\\\\?\\C:\\")
+     succeeding "/\\?\\C:\\" (Path "\\\\?\\C:\\")
+     succeeding "\\\\?\\C:\\\\\\folder\\\\" (Path "\\\\?\\C:\\folder\\")
 
   where failing x = parserTest parseAbsDir x Nothing
         succeeding x with = parserTest parseAbsDir x (Just with)
@@ -235,16 +255,17 @@ parseAbsDirSpec =
 parseRelDirSpec :: Spec
 parseRelDirSpec =
   do failing ""
-     -- failing "/" FIXME
-     -- failing "//" FIXME
-     -- succeeding "~/" (Path "~/") -- https://github.com/chrisdone/path/issues/19
-     -- failing "\\" FIXME
-     succeeding ".\\" (Path "")
-     succeeding ".\\.\\" (Path "")
+     failing "/"
+     failing "//"
+     failing "\\"
      failing "\\\\"
      failing "\\\\\\foo\\\\bar\\\\mu\\"
      failing "\\\\\\foo\\\\bar\\\\\\\\mu"
      failing "\\\\\\foo\\\\bar\\.\\\\mu"
+     failing "\\\\unchost\\share"
+     failing "\\\\?\\C:\\"
+     succeeding ".\\" (Path "")
+     succeeding ".\\.\\" (Path "")
      succeeding "..." (Path "...\\")
      succeeding "foo.bak" (Path "foo.bak\\")
      succeeding ".\\foo" (Path "foo\\")
@@ -269,10 +290,16 @@ parseAbsFileSpec =
      failing "\\"
      failing "\\\\"
      failing "\\\\\\foo\\\\bar\\\\mu\\"
-     -- succeeding "\\..." (Path "\\...") FIXME
-     -- succeeding "\\foo.txt" (Path "\\foo.txt") FIXME
-     -- succeeding "C:\\\\\\foo\\\\bar\\\\\\\\mu.txt" (Path "C:\\foo\\bar\\mu.txt") FIXME
-     -- succeeding "C:\\\\\\foo\\\\bar\\.\\\\mu.txt" (Path "C:\\foo\\bar\\mu.txt") FIXME
+     failing "\\..."
+     failing "\\foo.txt"
+     succeeding "C:\\\\\\foo\\\\bar\\\\\\\\mu.txt" (Path "C:\\foo\\bar\\mu.txt")
+     succeeding "C:\\\\\\foo\\\\bar\\.\\\\mu.txt" (Path "C:\\foo\\bar\\mu.txt")
+     succeeding "\\\\unchost\\share\\\\file.txt" (Path "\\\\unchost\\share\\file.txt")
+     succeeding "\\/unchost\\share\\\\file.txt" (Path "\\\\unchost\\share\\file.txt")
+     succeeding "\\\\unchost\\share\\.\\folder\\\\\\file.txt" (Path "\\\\unchost\\share\\folder\\file.txt")
+     succeeding "\\\\?\\C:\\file.txt" (Path "\\\\?\\C:\\file.txt")
+     succeeding "/\\?\\C:\\file.txt" (Path "\\\\?\\C:\\file.txt")
+     succeeding "\\\\?\\C:\\\\\\folder\\.\\\\file.txt" (Path "\\\\?\\C:\\folder\\file.txt")
 
   where failing x = parserTest parseAbsFile x Nothing
         succeeding x with = parserTest parseAbsFile x (Just with)
@@ -294,6 +321,8 @@ parseRelFileSpec =
      failing "\\\\\\foo\\\\bar\\\\mu\\"
      failing "\\\\\\foo\\\\bar\\\\\\\\mu"
      failing "\\\\\\foo\\\\bar\\.\\\\mu"
+     failing "\\\\unchost\\share\\\\file.txt"
+     failing "\\\\?\\C:\\file.txt"
      succeeding "a.." (Path "a..")
      succeeding "..." (Path "...")
      succeeding "foo.txt" (Path "foo.txt")
